from typing import Any, Awaitable, Callable

from .event import Event
from .signature import AsyncSignatureEvent, P, R, SignatureEvent

Method = Callable[..., Any]
Observer = Callable[..., Any]


def event(method: Callable[P, R]) -> SignatureEvent[P, R]:
    return SignatureEvent[P, R](method)


def async_event(method: Callable[P, Awaitable[R]]) -> AsyncSignatureEvent[P, R]:
    return AsyncSignatureEvent[P, R](method)


class BeforeEventMethodWrapper:
    def __init__(self, method: Method) -> None:
        self.method = method
        self.beforeEvent: Event[Any] = Event()

    def run(self, *args: Any, **kwargs: Any) -> None:
        self.beforeEvent(self.beforeEvent)
        self.method(*args, **kwargs)

    def subscribe(self, observer: Observer) -> "BeforeEventMethodWrapper":
        self.beforeEvent += observer
        return self

    def unsubscribe(self, observer: Observer) -> "BeforeEventMethodWrapper":
        self.beforeEvent -= observer
        return self

    __call__ = run
    __iadd__ = subscribe
    __isub__ = unsubscribe


def triggers_before_event(method: Method) -> BeforeEventMethodWrapper:
    return BeforeEventMethodWrapper(method)


class AfterEventMethodWrapper:
    def __init__(self, method: Method) -> None:
        self.method = method
        self.afterEvent: Event[Any] = Event()

    def run(self, *args: Any, **kwargs: Any) -> None:
        self.method(*args, **kwargs)
        self.afterEvent(self.afterEvent)

    def subscribe(self, observer: Observer) -> "AfterEventMethodWrapper":
        self.afterEvent += observer
        return self

    def unsubscribe(self, observer: Observer) -> "AfterEventMethodWrapper":
        self.afterEvent -= observer
        return self

    __call__ = run
    __iadd__ = subscribe
    __isub__ = unsubscribe


def triggers_after_event(method: Method) -> AfterEventMethodWrapper:
    return AfterEventMethodWrapper(method)


class AroundEventMethodWrapper:
    def __init__(self, method: Method) -> None:
        self.method = method
        self.beforeEvent: Event[Event[Any]] = Event()
        self.afterEvent: Event[Event[Any]] = Event()

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        self.beforeEvent(self.beforeEvent)
        self.method(*args, **kwargs)
        self.afterEvent(self.afterEvent)

    def subscribe(self, before: Observer, after: Observer) -> None:
        self.beforeEvent += before
        self.afterEvent += after


def triggers_beforeafter_events(method: Method) -> AroundEventMethodWrapper:
    return AroundEventMethodWrapper(method)



from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    ParamSpec,
    TypeVar,
    Union,
)

from .base import BaseEvent

P = ParamSpec("P")
R = TypeVar("R")


class SignatureEvent(Generic[P, R], BaseEvent):
    def __init__(self, method: Callable[P, R]) -> None:
        super().__init__()
        self._method = method

    def fire(self, *args: P.args, **kwargs: P.kwargs) -> R:
        self._fire(*args, **kwargs)
        return self._method(*args, **kwargs)

    __call__ = fire

    def __iadd__(
        self, subscribers: Union[Callable[P, Any], Iterable[Callable[P, Any]]]
    ) -> "SignatureEvent[P, R]":
        """Adds given `subscriber` and returns this Event"""
        self.append(subscribers)
        return self

    def __isub__(self, subscriber: Callable[P, Any]) -> "SignatureEvent[P, R]":
        """Removes given `subscriber` and returns this Event"""
        # only modify self if we're not currently firing
        # otherwise queue the unsubscription for processing after firing is done
        self.remove(subscriber)
        return self

    def append(self, subscribers: Union[Callable[P, Any], Iterable[Callable[P, Any]]]) -> None:
        super().append(subscribers)

    def remove(self, subscriber: Callable[P, Any]) -> None:
        super().remove(subscriber)

    def add(self, subscriber: Callable[P, Any]) -> Callable[[], Any]:
        return super().add(subscriber)


class AsyncSignatureEvent(Generic[P, R], BaseEvent):
    def __init__(self, method: Callable[P, Awaitable[R]]) -> None:
        super().__init__()
        self._method = method

    async def fire(self, *args: P.args, **kwargs: P.kwargs) -> R:
        for subscriber in self._iter():
            await subscriber(*args, **kwargs)

        return await self._method(*args, **kwargs)

    __call__ = fire

    def __iadd__(
        self, subscribers: Union[Callable[P, Awaitable[Any]], Iterable[Callable[P, Awaitable[Any]]]]
    ) -> "AsyncSignatureEvent[P, R]":
        """Adds given `subscriber` and returns this Event"""
        self.append(subscribers)
        return self

    def __isub__(self, subscriber: Callable[P, Awaitable[Any]]) -> "AsyncSignatureEvent[P, R]":
        """Removes given `subscriber` and returns this Event"""
        # only modify self if we're not currently firing
        # otherwise queue the unsubscription for processing after firing is done
        self.remove(subscriber)
        return self

    def append(
        self, subscribers: Union[Callable[P, Awaitable[Any]], Iterable[Callable[P, Awaitable[Any]]]]
    ) -> None:
        super().append(subscribers)

    def remove(self, subscriber: Callable[P, Awaitable[Any]]) -> None:
        super().remove(subscriber)

    def add(self, subscriber: Callable[P, Awaitable[Any]]) -> Callable[[], Any]:
        return super().add(subscriber)



import logging
from typing import Any, Callable, Generator, Iterable, Union

log = logging.getLogger(__name__)


class BaseEvent:
    @property
    def is_firing(self) -> bool:
        return self._currentFireCount > 0

    def __init__(self) -> None:
        self._subscribers: list[Callable[..., Any]] = []
        self._fireCount = 0
        self._currentFireCount = 0
        self._add_queue: list[Callable[..., Any]] = []
        self._remove_queue: list[Callable[..., Any]] = []

    def _fire(self, *args: Any, **kwargs: Any) -> Any:
        for subscriber in self._iter():
            subscriber(*args, **kwargs)

    def append(self, subscribers: Union[Callable[..., Any], Iterable[Callable[..., Any]]]) -> None:
        if callable(subscribers):
            subscribers = [subscribers]

        for subscriber in subscribers:
            if subscriber not in self._subscribers:
                if self.is_firing:
                    self._add_queue.append(subscriber)
                else:
                    self._subscribers.append(subscriber)

    def remove(self, subscriber: Callable[..., Any]) -> None:
        if self.is_firing:
            self._remove_queue.append(subscriber)
            return

        if subscriber not in self._subscribers:
            log.warning("Got unknown subscriber to remove from event")
            return

        self._subscribers.remove(subscriber)

    def add(self, subscriber: Callable[..., Any]) -> Callable[[], None]:
        """Same as `append` but returns a callable without arguments
        that can be used to unsubscribe the `subscriber`"""
        self.append(subscriber)

        def unsub() -> None:
            self.remove(subscriber)

        return unsub

    def _iter(self) -> Generator[Callable[..., Any], None, None]:
        # count the number of (recursive) fires currently happening
        self._currentFireCount += 1

        # execute all subscribers
        for subscriber in self._subscribers:
            # the handler might have got unsubscribed
            # inside one of the previous subscribers
            if subscriber not in self._remove_queue:
                # self._invoke(subscriber, value)
                yield subscriber

        # current fire cycle is done, uncount it
        self._currentFireCount -= 1

        # we're counting the number of fires (mostly for testing purposes)
        self._fireCount += 1

        # only if we're not still in a recursive fire situation
        if not self.is_firing:
            self._process_queues()

    def _process_queues(self) -> None:
        for subscriber in self._add_queue:
            self.append(subscriber)

        for subscriber in self._remove_queue:
            self.remove(subscriber)

        # reset processed queues
        self._add_queue = []
        self._remove_queue = []

    def __repr__(self) -> str:
        return f"Event(id={id(self)}, len={len(self)})"

    def __eq__(self, other: Any) -> bool:
        return id(self) == id(other)

    def __len__(self) -> int:
        return len(self._subscribers)

    def __contains__(self, subscriber: Callable[..., Any]) -> bool:
        return subscriber in self._subscribers



from typing import Any, Awaitable, Callable, Generic, Iterable, TypeVar, Union

from .base import BaseEvent

T = TypeVar("T")


class AsyncEvent(Generic[T], BaseEvent):
    async def fire(self, value: T) -> None:
        for subscriber in self._iter():
            await subscriber(value)

    __call__ = fire

    def __iadd__(
        self,
        subscribers: Union[Callable[[T], Awaitable[Any]], Iterable[Callable[[T], Awaitable[Any]]]],
    ) -> "AsyncEvent[T]":
        """Adds given `subscriber` and returns this Event"""
        self.append(subscribers)
        return self

    def __isub__(self, subscriber: Callable[[T], Awaitable[Any]]) -> "AsyncEvent[T]":
        """Removes given `subscriber` and returns this Event"""
        # only modify self if we're not currently firing
        # otherwise queue the unsubscription for processing after firing is done
        self.remove(subscriber)
        return self

    def append(
        self,
        subscribers: Union[Callable[[T], Awaitable[Any]], Iterable[Callable[[T], Awaitable[Any]]]],
    ) -> None:
        super().append(subscribers)

    def remove(self, subscriber: Callable[[T], Awaitable[Any]]) -> None:
        super().remove(subscriber)

    def add(self, subscriber: Callable[[T], Awaitable[Any]]) -> Callable[[], Any]:
        return super().add(subscriber)



from .async_event import AsyncEvent
from .decorators import (
    async_event,
    event,
    triggers_after_event,
    triggers_before_event,
    triggers_beforeafter_events,
)
from .event import Event

__all__ = [
    "async_event",
    "AsyncEvent",
    "event",
    "Event",
    "triggers_after_event",
    "triggers_before_event",
    "triggers_beforeafter_events",
]
